import torch
from modules_ot.solvers import DRPASGDSolver, LogSinkhornSolver  

class OTProblem:
    def __init__(self, dim, draw_mu, nu_points, nu_weights, epsilon=0, max_cost=1, theoretical_g_opt=None):
        self.dim = dim
        self.draw_mu = draw_mu 
        self.nu_points = nu_points  
        self.nu_weights = nu_weights 
        self.theoretical_g_opt = theoretical_g_opt  
        self.solver = DRPASGDSolver()
        self.g_approx = None
        self.f_approx = None
        self.epsilon = epsilon
        self.max_cost = max_cost

    def print_setting(self): 
        print(f"Dimension: {self.dim}")
        print(f"Source measure: {self.draw_mu}")
        print(f"Target measure's has : {self.nu_points.shape[0]} points")
        # check if the weights are uniform
        if torch.all(torch.eq(self.nu_weights, self.nu_weights[0])):
            print(f"Target measure's weights are uniform: {self.nu_weights[0]}")
        else:
            print(f"Target measures's weights are between {self.nu_weights.min()} and {self.nu_weights.max()}")
        print(f"Theoretical g_opt: {self.theoretical_g_opt}")
        print(f"Approximated g_opt: {self.g_approx}")
        print(f"Solver: {self.solver}")
        print(f"Epsilon: {self.epsilon}")
        print(f"Max cost: {self.max_cost}")


    def _select_solver(self, solver_name):
        if solver_name == "drpasgd":
            self.solver =  DRPASGDSolver()
        elif solver_name == "log_sinkhorn":
            self.solver = LogSinkhornSolver()
        else:
            raise ValueError(f"Solver {solver_name} not implemented")

    def solve(self):
        print("Solving...")
        g_t, g_bar, g_hist, g_bar_hist, g_weighted_hist = self.solver.solve(self.dim, self.draw_mu, self.nu_points, self.nu_weights, self.max_cost)
        self.g_approx = g_bar
        print("Done.")
        return g_hist, g_bar_hist, g_weighted_hist

    def change_solver_parameters(self, params):
        self.solver.change_parameters(params)

    def cost_function(self, x, y):
        return 0.5 * torch.sum((x - y) ** 2, dim=-1)

    def c_transform(self, x, g):
        costs = self.cost_function(x, self.nu_points)
        return torch.min(costs - g)

    def transport_point(self, x, g):
        v = self.cost_function(self.nu_points, x) - g
        i_min = torch.argmin(v)
        return self.nu_points[i_min]

    def eot_transport_point(self, x, g):
        v = self.cost_function(self.nu_points, x) - g
        weights = torch.nn.functional.softmax(-v/self.epsilon, dim=0)
        weighted_Y = self.nu_points.T*weights 
        return torch.sum(weighted_Y, axis=1)

    def transport_error_point(self, x, g):
        theoretical_y = self.transport_point(x, self.theoretical_g_opt)
        approx_y = self.transport_point(x, g)
        return torch.norm(theoretical_y - approx_y)**2

    def transport_error(self, g):
        error = 0
        nb_montecarlo = 10000
        for _ in range(nb_montecarlo):
            x = self.draw_mu(1, self.dim)
            error += self.transport_error_point(x, g)
        return error / nb_montecarlo
    
    def cost_dual(self, g):
        nb_loop = 10
        mc_cost = 0
        for _ in range(nb_loop):
            nb_montecarlo = 100000
            # Draw all samples at once
            X = self.draw_mu(nb_montecarlo, self.dim)

            # Compute the squared Euclidean distance 0.5 * ||x - y||^2 for each x from each y
            # X shape: (nb_montecarlo, 1, dim), nu_points shape: (1, num_nu_points, dim)
            distances = 0.5 * (X.unsqueeze(1) - self.nu_points.unsqueeze(0)).pow(2).sum(dim=2)

            # Subtract g from distances and take the minimum along the nu_points dimension
            # Ensure g is shaped (num_nu_points,) for proper broadcasting
            min_costs = torch.min(distances - g, dim=1)[0]

            # Compute the overall Monte Carlo estimate
            mc_cost += torch.mean(min_costs + torch.sum(g * self.nu_weights))
        mc_cost /= nb_loop
        return mc_cost
    
    def dual_error(self, g):
        th_cost = self.cost_dual(self.theoretical_g_opt)
        approx_cost = self.cost_dual(g)
        return torch.abs(th_cost - approx_cost)

